# Copyright 2016 Yanis Guenane <yguenane@redhat.com>
# Author: Yanis Guenane <yguenane@redhat.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from lecm import utils
from OpenSSL import crypto

import datetime
import logging
import os
import requests
import socket
import subprocess

LOG = logging.getLogger(__name__)

_INTERMEDIATE_CERTIFICATE_URL = \
    'https://letsencrypt.org/certs/lets-encrypt-r3-cross-signed.pem'

_STAGING_URL = \
    'https://acme-staging.api.letsencrypt.org'


class Certificate(object):

    def __init__(self, conf):
        self.name = conf.get('name')
        self.path = conf.get('path')
        self.type = conf.get('type', 'RSA')
        self.size = conf.get('size', 4096)
        self.digest = conf.get('digest', 'sha256')
        self.version = conf.get('version', 3)
        self.environment = conf.get('environment', 'production')
        self.subjectAltName = self.normalize_san(conf.get('subjectAltName'))
        self.account_key_name = conf.get('account_key_name',
                                         'account_%s.key' % socket.getfqdn())
        self.remaining_days = conf.get('remaining_days', 10)
        self.days_before_expiry = self.get_days_before_expiry()
        self.service_name = conf.get('service_name', 'httpd')
        self.service_provider = conf.get('service_provider', 'systemd')

        self.subject = {
          'C': conf.get('countryName'),
          'ST': conf.get('stateOrProvinceName'),
          'L': conf.get('localityName'),
          'O': conf.get('organizationName'),
          'OU': conf.get('organizationalUnitName'),
          'CN': conf.get('commonName'),
          'emailAddress': conf.get('emailAddress'),
        }

        if self.subject['CN'] is None:
            self.subject['CN'] = self.name

        if self.subjectAltName is None:
            self.subjectAltName = 'DNS:%s' % self.name

    def normalize_san(self, san):

        # If an array of SAN is passed in the configuration file
        #
        # certificates:
        #   my.example.org:
        #     subjectAltName:
        #       - my.example.org
        #       - my1.example.org
        #
        # return : DNS:my.example.org,DNS:my1.example.org
        if isinstance(san, list):
            san_string = 'DNS:%s' % ',DNS:'.join(san)

        # If a string of SAN is passed but without the proper format
        #
        # certificates:
        #   my.example.org:
        #     subjectAltName: my.example.org,my1.example.org
        #
        # return : DNS:my.example.org,DNS:my1.example.org
        elif san and not san.startswith('DNS:'):
            san_string = 'DNS:%s' % ',DNS:'.join(san.split(','))
        else:
            san_string = 'DNS:%s' % self.name

        return san_string

    def _create_filesystem(self):
        _FOLDERS = ['csr', 'challenges', 'pem', 'private', 'certs']

        for folder in _FOLDERS:
            LOG.debug('[global] Ensure path exist: %s/%s' %
                      (self.path, folder))
            if not os.path.exists('%s/%s' % (self.path, folder)):
                os.makedirs('%s/%s' % (self.path, folder))
                utils.enforce_selinux_context(self.path)

    def _get_intermediate_certificate(self):
        certificate_name = os.path.basename(_INTERMEDIATE_CERTIFICATE_URL)
        if not os.path.exists('%s/pem/%s' % (self.path, certificate_name)):
            certificate = requests.get(_INTERMEDIATE_CERTIFICATE_URL).text

            LOG.info('[global] Getting intermediate certificate PEM file: %s' %
                     certificate_name)
            if not os.path.exists('%s/pem/%s' % (self.path, certificate_name)):
                with open('%s/pem/%s' % (self.path, certificate_name), 'w') as f:
                    f.write(certificate)

    def _create_account_key(self):
        account_key = crypto.PKey()

        if self.type == 'RSA':
            crypto_type = crypto.TYPE_RSA
        else:
            crypto_type = crypto.TYPE_DSA

        try:
            LOG.info('[global] Generating account key: %s \
                     (type: %s, size: %s)' %
                     (self.account_key_name, self.type, self.size))
            account_key.generate_key(crypto_type, self.size)
        except (TypeError, ValueError):
            raise

        try:
            LOG.debug('[global] Writting account key: %s/private/%s' %
                      (self.path, self.account_key_name))
            accountkey_file = os.open('%s/private/%s' %
                                      (self.path, self.account_key_name),
                                      os.O_WRONLY | os.O_CREAT | os.O_TRUNC,
                                      0o600)
            os.write(accountkey_file,
                     crypto.dump_privatekey(crypto.FILETYPE_PEM, account_key))
            os.close(accountkey_file)
        except IOError:
            try:
                os.remove('%s/private/%s.key' %
                          (self.path, self.account_key_name))
            except OSError:
                pass
            raise

    def _create_private_key(self):
        private_key = crypto.PKey()

        if self.type == 'RSA':
            crypto_type = crypto.TYPE_RSA
        else:
            crypto_type = crypto.TYPE_DSA

        try:
            LOG.info('[%s] Generating private key (type: %s, size: %s)' %
                     (self.name, self.type, self.size))
            private_key.generate_key(crypto_type, self.size)
        except (TypeError, ValueError):
            raise

        try:
            LOG.debug('[%s] Writting private key: %s/private/%s.key' %
                      (self.name, self.path, self.name))
            privatekey_file = os.open('%s/private/%s.key' %
                                      (self.path, self.name),
                                      os.O_WRONLY | os.O_CREAT | os.O_TRUNC,
                                      0o600)
            os.write(privatekey_file,
                     crypto.dump_privatekey(crypto.FILETYPE_PEM, private_key))
            os.close(privatekey_file)
        except IOError:
            try:
                os.remove('%s/private/%s.key' % (self.path, self.name))
            except OSError:
                pass
            raise

    def _create_csr(self):
        LOG.info('[%s] Generating CSR' % self.name)
        req = crypto.X509Req()
        LOG.debug('[%s] Attaching Certificate Version to CSR: %s' %
                  (self.name, self.version))
        req.set_version(self.version)
        subject = req.get_subject()

        for (key, value) in self.subject.items():
            if value is not None:
                LOG.debug('[%s] Attaching %s to CSR: %s' %
                          (self.name, key, value))
                setattr(subject, key, value)

        LOG.info('[%s] Attaching SAN extention: %s' %
                 (self.name, self.subjectAltName))

        try:
            req.add_extensions([crypto.X509Extension(
                bytes('subjectAltName', 'utf-8'), False,
                bytes(self.subjectAltName, 'utf-8')
            )])
        except TypeError:
            req.add_extensions([crypto.X509Extension('subjectAltName', False,
                                                     self.subjectAltName)])

        LOG.debug('[%s] Loading private key: %s/private/%s.key' %
                  (self.name, self.path, self.name))
        privatekey_content = open('%s/private/%s.key' %
                                  (self.path, self.name)).read()

        privatekey = crypto.load_privatekey(crypto.FILETYPE_PEM,
                                            privatekey_content)

        LOG.info('[%s] Signing CSR' % self.name)
        req.set_pubkey(privatekey)
        req.sign(privatekey, self.digest)

        LOG.debug('[%s] Writting CSR: %s/csr/%s.csr' %
                  (self.name, self.path, self.name))
        csr_file = open('%s/csr/%s.csr' % (self.path, self.name), 'w')
        csr_file.write((crypto.dump_certificate_request(crypto.FILETYPE_PEM,
                                                        req)).decode('utf-8'))
        csr_file.close()

    def _create_certificate(self):
        LOG.info('[%s] Retrieving certificate from Let''s Encrypt Server' %
                 self.name)
        command = 'acme-tiny --account-key %s/private/%s --csr %s/csr/%s.csr \
                  --acme-dir %s/challenges/%s' % (self.path,
                                                  self.account_key_name,
                                                  self.path, self.name,
                                                  self.path, self.name)

        if self.environment == 'staging':
            LOG.info('[%s] Using Let''s Encrypt staging API: %s' %
                     (self.name, _STAGING_URL))
            command = '%s --ca %s' % (command, _STAGING_URL)

        cert_file_f = open('%s/certs/%s.crt.new' % (self.path, self.name), 'w')

        p = subprocess.Popen(command.split(), stdout=cert_file_f,
                             stderr=subprocess.PIPE)
        out, err = p.communicate()

        if p.returncode != 0:
            LOG.error('[%s] %s' % (self.name, err))
            os.remove('%s/certs/%s.crt.new' % (self.path, self.name))
            return False
        else:
            LOG.debug('[%s] Writting certificate: %s/certs/%s.crt' %
                      (self.name, self.path, self.name))
            os.rename('%s/certs/%s.crt.new' % (self.path, self.name),
                      '%s/certs/%s.crt' % (self.path, self.name))

            LOG.debug('[%s] Concatenating certificate with intermediate pem: \
                      %s/pem/%s.pem' % (self.name, self.path, self.name))
            self._get_intermediate_certificate()
            pem_filename = os.path.basename(_INTERMEDIATE_CERTIFICATE_URL)
            filenames = ['%s/certs/%s.crt' % (self.path, self.name),
                         '%s/pem/%s' % (self.path, pem_filename)]
            with open('%s/pem/%s.pem' % (self.path, self.name), 'w') as f:
                for fname in filenames:
                    with open(fname) as infile:
                        f.write(infile.read())
        return True

    def get_days_before_expiry(self):
        try:
            x509_content = open('%s/pem/%s.pem' %
                                (self.path, self.name)).read()
        except IOError:
            return 'N/A'
        x509 = crypto.load_certificate(crypto.FILETYPE_PEM, x509_content)

        notAfter = x509.get_notAfter()[:-1]
        notAfter_datetime = datetime.datetime.strptime(
            notAfter.decode('utf-8'),
            '%Y%m%d%H%M%S'
        )
        now_datetime = datetime.datetime.now()

        return (notAfter_datetime - now_datetime).days

    def generate(self):

        self._create_filesystem()

        # Ensure there is no left-over from previous setup
        #
        try:
            LOG.info('[%s] Removing older files (if any)' % self.name)
            os.remove('%s/private/%s.key' % (self.path, self.name))
            os.remove('%s/csr/%s.csr' % (self.path, self.name))
            os.rmdir('%s/challenges/%s' % (self.path, self.name))
            os.remove('%s/certs/%s.key' % (self.path, self.name))
        except OSError:
            pass

        if not os.path.exists('%s/private/%s' %
                              (self.path, self.account_key_name)):
            self._create_account_key()
        self._create_private_key()
        self._create_csr()
        os.makedirs('%s/challenges/%s' % (self.path, self.name))
        self._create_certificate()

    def renew(self):
        self._create_csr()
        self._create_certificate()

    def reload_service(self):
        utils.reload_service(self.service_name, self.service_provider)



# Copyright 2016 Yanis Guenane <yguenane@redhat.com>
# Author: Yanis Guenane <yguenane@redhat.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from lecm import exceptions

import logging
import os
import yaml


LOG = logging.getLogger(__name__)

_FIELDS = ['type', 'size', 'digest', 'version', 'subjectAltName',
           'countryName', 'stateOrProvinceName', 'localityName',
           'organizationName', 'organizationUnitName', 'commonName',
           'emailAddress', 'account_key_name', 'path', 'remaining_days',
           'service_name', 'service_provider', 'environment']


def check_configuration_file_existence(configuration_file_path=None):
    """Check if the configuration file is present."""

    if configuration_file_path:
        if not os.path.exists(configuration_file_path):
            raise exceptions.ConfigurationExceptions(
                      'File %s does not exist' % configuration_file_path
                  )
        file_path = configuration_file_path
    elif os.getenv('LECM_CONFIGURATION'):
        if not os.path.exists(os.getenv('LECM_CONFIGURATION')):
            raise exceptions.ConfigurationExceptions(
                      'File %s does not exist' %
                      os.getenv('LECM_CONFIGURATION')
                  )
        file_path = os.getenv('LECM_CONFIGURATION')
    else:
        if not os.path.exists('/etc/lecm.conf'):
            raise exceptions.ConfigurationExceptions(
                'File /etc/lecm.conf does not exist (you could specify an '
                'alternate location using --conf)'
            )
        file_path = '/etc/lecm.conf'

    LOG.debug('Configuration file used: %s' % file_path)
    return file_path


def load_configuration(conf):
    """Load the lecm configuration file."""

    file_path = check_configuration_file_existence(conf.get('file_path'))

    try:
        file_path_content = open(file_path, 'r').read()
    except IOError as exc:
        raise exceptions.ConfigurationExceptions(exc)

    try:
        conf = yaml.load(file_path_content)
    except yaml.YAMLError as exc:
        raise exceptions.ConfigurationExceptions(exc)

    return conf


def expand_configuration(configuration):
    """Fill up certificates with defaults."""

    certificates = {}
    for name, parameters in configuration['certificates'].items():
        if not isinstance(parameters, dict):
            parameters = {}
        parameters['name'] = name
        for field in _FIELDS:
            if field not in parameters.keys() or parameters[field] is None:
                if field in configuration:
                    parameters[field] = configuration[field]
        certificates[name] = parameters

    return certificates



# Copyright 2016 Yanis Guenane <yguenane@redhat.com>
# Author: Yanis Guenane <yguenane@redhat.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from lecm import certificate
from lecm import configuration
from lecm import lists
from lecm import parser
from lecm import utils

import logging
import os
import sys


def should_reload(cert, global_configuration):

    if cert.service_name != global_configuration.get('service_name', 'httpd'):
        return True

    return False


def main():

    options = parser.parse()

    if isinstance(options, int):
        sys.stderr.write(
            'USAGE: lecm [--generate,--renew,--list,--list-details]\n'
        )
        return 1

    if options.debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    _CONF = {}
    if options.conf:
        _CONF['file_path'] = options.conf

    global_configuration = configuration.load_configuration(_CONF)
    certificates = configuration.expand_configuration(global_configuration)
    certificates = utils.filter_certificates(options.items, certificates)

    if options.list:
        lists.list(certificates)
    elif options.list_details:
        lists.list_details(certificates)
    else:
        noop_holder = {}
        certs = []
        certs_w_service_name = []
        services_to_restart = []

        for name, parameters in certificates.items():
            cert = certificate.Certificate(parameters)
            if options.generate:
                if options.noop:
                    if not os.path.exists('%s/pem/%s.pem' %
                                          (cert.path, cert.name)):
                        noop_holder[name] = parameters
                else:
                    if not os.path.exists('%s/pem/%s.pem' %
                                          (cert.path, cert.name)) or \
                       options.force:
                        cert.generate()
                        certs.append(cert)
                        if 'service_name' in parameters:
                            certs_w_service_name.append(cert)
                            if not isinstance(parameters['service_name'], list):  # noqa
                                services_to_restart.append(
                                    [parameters['service_name']]
                                )
                            else:
                                services_to_restart.append(
                                    parameters['service_name']
                                )

            elif options.renew:
                if options.noop:
                    if isinstance(cert.days_before_expiry, int) and \
                       cert.days_before_expiry <= cert.remaining_days:
                        noop_holder[name] = parameters
                else:
                    if (isinstance(cert.days_before_expiry, int) and
                       cert.days_before_expiry <= cert.remaining_days) or \
                       options.force:
                        cert.renew()
                        certs.append(cert)
                        if 'service_name' in parameters:
                            certs_w_service_name.append(cert)
                            if not isinstance(parameters['service_name'], list):  # noqa
                                services_to_restart.append(
                                    [parameters['service_name']]
                                )
                            else:
                                services_to_restart.append(
                                    parameters['service_name']
                                )

        if len(certs) != len(certs_w_service_name):
            services_to_restart.append(
                global_configuration.get('service_name', ['httpd'])
            )

        if certs and not options.noop:
            utils.reload_service(
                list(set(sum(services_to_restart, []))),
                global_configuration.get('service_provider', 'systemd')
            )
        if options.noop:
            lists.list(noop_holder)


if __name__ == "__main__":
    main()



# Copyright 2016 Yanis Guenane <yguenane@redhat.com>
# Author: Yanis Guenane <yguenane@redhat.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from prettytable import PrettyTable
from OpenSSL import crypto

import copy
import logging
import os
import platform
import subprocess

LOG = logging.getLogger(__name__)


def output_informations(data):

    x = PrettyTable()
    for column in data:
        x.add_column(column[0], column[1])
    print(x)


def filter_certificates(items, certificates):

    certificates_to_return = copy.deepcopy(certificates)
    if isinstance(items, list):
        for name in certificates.keys():
            if name not in items:
                del certificates_to_return[name]

    return certificates_to_return


def enforce_selinux_context(output_directory):

    if platform.dist()[0] in ['fedora', 'centos', 'redhat']:
        if os.path.exists('/sbin/semanage'):
            FNULL = open(os.devnull, 'w')

            # Set new selinux so it is persistent over reboot
            command = 'semanage fcontext -a -t cert_t %s(/.*?)' % (
                output_directory
            )
            p = subprocess.Popen(command.split(), stdout=FNULL,
                                 stderr=subprocess.STDOUT)
            p.wait()

            # Ensure file have the right context applied
            command = 'restorecon -Rv %s' % output_directory
            p = subprocess.Popen(command.split(), stdout=FNULL,
                                 stderr=subprocess.STDOUT)
            p.wait()


def reload_service(service_name, service_provider):

    if service_name:
        if not isinstance(service_name, list):
            service_name = [service_name]

        for service in service_name:
            LOG.info('Reloading service specified: %s' % service)
            if service_provider == 'sysv':
                command = 'service %s reload' % service
            else:
                command = 'systemctl reload %s' % service
            p = subprocess.Popen(command.split())
            p.wait()


def get_subjectaltname(certificate):
    """Return subjectAltName associated with certificate. """

    return certificate.get_extension(6)._subjectAltNameString()


def get_environment(certificate):
    """Return environment associated with certificate. """

    STAGING_ENV = "Fake LE Intermediate X1"

    for component in certificate.get_issuer().get_components():
        if component[0] == 'CN':
            if component[1] == STAGING_ENV:
                return 'staging'
            else:
                return 'production'

    return None


def is_sync(certificate):
    """Return true or false if certificate and definition are in sync,

       Certificate and definitions are said to be in sync if the following
       parameters match:

         * Issuer CN
         * SubjectAltName"""

    original_certificate = '%s/pem/%s.pem' % (certificate['path'],
                                              certificate['name'])

    if not os.path.exists(original_certificate):
        return False

    buf = open(original_certificate).read()
    pem = crypto.load_certificate(crypto.FILETYPE_PEM, buf)

    cur_environment = get_environment(pem)
    cur_subjectaltname = get_subjectaltname(pem)
    cert_san = certificate.get('subjectAltName', [certificate.get('name')])

    if cur_environment != certificate.get('environment', 'production') or \
        cur_subjectaltname != \
            'DNS:%s' % ', DNS:'.join(sorted(cert_san)):
        return False

    return True



# Copyright 2016 Yanis Guenane <yguenane@redhat.com>
# Author: Yanis Guenane <yguenane@redhat.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from lecm.version import __version__

import argparse


def parse():

    parser = argparse.ArgumentParser(
        description='Let''s Encrypt Certificate Manager',
        epilog='lecm [--generate,--renew,--list,--list-details]'
    )

    parser.add_argument('-v', '--version', action='version',
                        version='%(prog)s ' + __version__)
    parser.add_argument('--debug', action='store_true',
                        help='Display DEBUG information level')

    parser.add_argument('--noop', action='store_true',
                        help='Proceed in noop mode')
    parser.add_argument('--conf', help='Path to configuration file')
    parser.add_argument('--items', action='append', nargs='*',
                        help='Limit the item to apply the action to')
    parser.add_argument('-l', '--list', action='store_true',
                        help='List the lecm configured certificates')
    parser.add_argument('-ld', '--list-details', action='store_true',
                        help='List the lecm configured certificates(details)')

    parser.add_argument('--generate', action='store_true',
                        help='Generate Let''s Encrypt SSL Certificates')
    parser.add_argument('--renew', action='store_true',
                        help='Renew already generated SSL Certificates')
    parser.add_argument('--force', action='store_true',
                        help='Force renewal or reneration of the SSL \
                              Certificates')
    options = parser.parse_args()

    if len([value for value in vars(options).values() if value]) == 0:
        return 1

    normalize_items_parameter(options)

    return options


def normalize_items_parameter(options):
    """The items parameters can have differents form based on how it
       was passed as an input

       lecm --generate --items my.example.com,my2.example.com
       lecm --generate --items my.example.com my2.example.com
       lecm --generate --items my.example.com --items my2.example.com

       This method aims to provide a plain array witch each element being
       a items itself
    """

    if not isinstance(options.items, list):
        return

    final_items = []
    for items in options.items:
        for item in items:
            if ',' in item:
                final_items += item.split(',')
            else:
                final_items.append(item)

    options.items = final_items



# Copyright 2016 Yanis Guenane <yguenane@redhat.com>
# Author: Yanis Guenane <yguenane@redhat.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from lecm import certificate
from lecm import utils

import os


def list(certificates):
    result = [['In Sync', []],
              ['Item', []],
              ['Status', []],
              ['subjectAltName', []],
              ['Location', []],
              ['Days', []]]
    for name, parameters in certificates.items():
        cert = certificate.Certificate(parameters)

        result[0][1].append(utils.is_sync(parameters))
        result[1][1].append(cert.name)
        if os.path.exists('%s/pem/%s.pem' % (cert.path, cert.name)):
            result[2][1].append('Generated')
        else:
            result[2][1].append('Not-Generated')
        result[3][1].append(cert.subjectAltName)
        result[4][1].append('%s/pem/%s.pem' % (cert.path, cert.name))
        result[5][1].append(cert.days_before_expiry)

    utils.output_informations(result)


def list_details(certificates):
    result = [['In Sync', []],
              ['Item', []],
              ['Status', []],
              ['subjectAltName', []],
              ['emailAddress', []],
              ['Environment', []],
              ['Location', []],
              ['Type', []],
              ['Size', []],
              ['Digest', []],
              ['Days', []]]
    for name, parameters in certificates.items():
        cert = certificate.Certificate(parameters)

        result[0][1].append(utils.is_sync(parameters))
        result[1][1].append(cert.name)
        if os.path.exists('%s/pem/%s.pem' % (cert.path, cert.name)):
            result[2][1].append('Generated')
        else:
            result[2][1].append('Not-Generated')
        result[3][1].append(cert.subjectAltName)
        result[4][1].append(cert.subject['emailAddress'])
        result[5][1].append(cert.environment)
        result[6][1].append('%s/pem/%s.pem' % (cert.path, cert.name))
        result[7][1].append(cert.type)
        result[8][1].append(cert.size)
        result[9][1].append(cert.digest)
        result[10][1].append(cert.days_before_expiry)

    utils.output_informations(result)
